#! perl -w

# Licensed Materials - Property of IBM 
# (c) Copyright IBM Corporation 1991, 2005 All Rights Reserved. 


# rtcomp.pl - Java compilation controller
#
# - produces DEP files (from import statements)
# - converts Java-compiler error messages into toolset-readable form.

# Force autoflush for nice error messages

select STDERR; $| = 1;
select STDOUT; $| = 1;

# Get script directory.

$scripts = $0;
$scripts =~ s{\\}{/}g;
$scripts =~ s{/rtcomp.pl$}{};

# load in the error handler routines
require "$scripts/rterror.pl";

# Silence perl: other uses are in rterror.pl.
@config::handlers  = ();
$rterror::maxError = $rterror::maxError;

# defaults
$spacify_mode = 'dq';
$vendor       = 'generic';

&get_args();

print STDERR "!> Compiling ", $class_name, "\n";

&run_cmd();

&handle_verbose_output();
&check_output_dependencies(); # warn if multiple files output

@config::handlers =
(
	[ '^(.*?):(\d+):(?:[:\d]+:)?\s+warning:\s+(.*)',
		sub { rterror::action_print( $1, $2, $3, 1 ); } ],
	[ '^(.*?):(\d+):(?:[:\d]+:)?\s+(.*)',
		sub { rterror::action_print( $1, $2, $3, 0 ); } ]
);

print @errors;

&rterror::handle_errors();

&exit_if_error();

&write_cmk_file(); # included by the Makefile for compilation dependencies
&write_ctg_file(); # timestamp guaranteed to be up-to-date for the Makefile

##
## Subroutines
##

sub get_args()
{
	# Get the arguments

	while( $#ARGV >= 0 )
	{
		$arg = shift @ARGV;
		if   ( $arg eq '-vendor'  ) { $vendor       = shift @ARGV; }
		elsif( $arg eq '-spacify' ) { $spacify_mode = shift @ARGV; }
		elsif( $arg eq '-target'  ) { $ctg_file     = shift @ARGV; }
		elsif( $arg eq '--'       ) { last; }
		else
		{
			&rterror::print_error( '', "Unknown rtcomp.pl option '$arg'", 1 );
		}
	}

	# first, let's make all the slashes go one way
	$ctg_file =~ s{\\}{/}g;

	$java_file = $ctg_file;
	$java_file =~ s{^RTbuild/(.*)\.ctg$}{$1.java};

	$cmk_file = $ctg_file;
	$cmk_file =~ s{\.ctg$}{.cmk};

	# get the class name
	$class_name = $ctg_file;
	$class_name =~ s{^RTbuild/(.*)\.ctg$}{$1};
	$class_name =~ s{/}{.}g;
}

sub run_cmd()
{
	$user_specified_verbose = 0;
	$verbose_flag           = ' -verbose';

	@command = ();
	foreach $arg ( @ARGV )
	{
		if( $arg =~ /-verbose/ )
		{
			$user_specified_verbose = 1;
		}
		elsif( $arg =~ /-noverbose/ )
		{
			$user_specified_verbose = -1;
			next; # we have to ignore this
		}
		$arg =~ s{\$\((\w+)\)}{$ENV{$1} || ''}eg;
		$arg =~ s{^.*\s.*$}{"$&"};
		push @command, $arg;
	}
	$command = join( ' ', @command );

	$OS          = $ENV{'OS'}          || '';
	$ROSERT_HOST = $ENV{'ROSERT_HOST'} || '';

	$errRedirect = ( $ROSERT_HOST eq 'win32' && $OS !~ m{^Windows_NT$}i )
							? ''
							: ' 2>&1';

	open OPEN, "$command$verbose_flag$errRedirect|";

	@errors = <OPEN>;

	close OPEN;
	$retcode = $?;
}

sub exit_if_error()
{
	if( $rterror::worstError < $rterror::maxError )
	{
		if( $rterror::worstError == 0 )
		{
			exit 1;
		}
	}
	elsif( $retcode & 0xff )
	{
		&rterror::print_error( '', "command [$command] caught signal [$retcode]", 0 );
		exit 1;
	}
	elsif( $retcode )
	{
		$retcode >>= 8;
		&rterror::print_error( '', "command [$command] returned error [$retcode]", 0 );
		exit 1;
	}
}

sub write_ctg_file()
{
	if( ! open CTGFILE, ">$ctg_file" )
	{
		&rterror::print_error( '', "Cannot write to '$ctg_file'", 0 );
		exit 1;
	}

	binmode( CTGFILE );
	print CTGFILE "# ", $ctg_file, "\n\n",
				  "# The timestamp of this file is used by RTcompile.mk\n",
				  "# to determine when to recompile.\n";
	close CTGFILE;
}

sub write_cmk_file()
{
	if( ! open DEPFILE, ">$cmk_file" )
	{
		&rterror::print_error( '', "Cannot write to '$cmk_file'", 0 );
		exit 1;
	}

	binmode( DEPFILE );

	$depvar = $class_name;
	$depvar =~ s{_}{__}g;
	$depvar =~ s{\.}{_}g;

	print DEPFILE "# ", $cmk_file, "\n\n",
				  $depvar, "_DEPS_J =";

	if( $spacify_mode eq 'dq' )
	{
		foreach ( @in_deps_list )
		{
			s{^.*\s.*$}{"$&"};
		}
	}
	elsif( $spacify_mode eq 'bs' )
	{
		foreach ( @in_deps_list )
		{
			s{(\s)}{\\$1}g;
		}
	}

	foreach ( @in_deps_list )
	{
		print DEPFILE " \\\n\t", $_;
	}

	print DEPFILE "\n";
	close DEPFILE;
}

sub add_input_dependency(*)
{
	my( $file ) = @_;

	# make all the slashes go one way
	$file =~ s{\\}{/}g;

	# remove leading "./"
	$file =~ s{^(\./)+}{};

	# Replace dependencies on Class$Nested.class into dependencies
	# on Class.class to avoid confusing make with the $ character.
	# All compilers are likely to write Class.class at the same time
	# as they would write Class$Nested.class.
	$file =~ s{\$.*\.class$}{.class}g;

	if( ! $in_deps_map{ $file } )
	{
		$in_deps_map{ $file } = 1;
		push @in_deps_list, $file;
	}
}

sub add_output_dependency(*)
{
	my( $file ) = @_;

	# make all the slashes go one way
	$file =~ s{\\}{/}g;

	if( ! $out_deps_map{ $file } )
	{
		$out_deps_map{ $file } = 1;
		push @out_deps_list, $file;
	}
}

sub ignore
{
}

sub handle_verbose_output()
{
	@verbose_handlers = # for JDK 1.3
	(
		[ '\[parsing started (.*)\]', sub { &add_input_dependency( $1 ); } ],
		[ '\[loading (.*)\(.*\)\]',   sub { &add_input_dependency( $1 ); } ],
		[ '\[loading ([^\(\)]*)\]',   sub { &add_input_dependency( $1 ); } ],
		[ '\[read (.*)\(.*\)\]',      sub { &add_input_dependency( $1 ); } ],
		[ '\[read ([^\(\)]*)\]',      sub { &add_input_dependency( $1 ); } ],
		[ '\[parsing completed .*\]', sub { &ignore;  } ],
		[ '\[checking .*\]',          sub { &ignore;  } ],
		[ '\[write (.*)\]',           sub { &add_output_dependency( $1 ); } ],
		[ '\[wrote (.*)\]',           sub { &add_output_dependency( $1 ); } ],
		[ '\[total .*\]',             sub { &ignore;  } ]
	);

	foreach $line ( @errors )
	{
		foreach $handler ( @verbose_handlers )
		{
			( $pattern, $action ) = @$handler;

			if( $line =~ /$pattern\n/ )
			{
				&$action;
				if( $user_specified_verbose <= 0 )
				{
					$line = '';
				}
			}
		}
	}
}

sub check_output_dependencies()
{
	# Java compiler should only produce one class.
	# Otherwise the makefile will get confused and/or run sub-optimally.

	if( $#out_deps_map > 0 )
	{
		&rterror::print_error( '',
							   'Multiple output classes detected ['
								   . join( ',', @out_deps_list )
								   . ']',
							   1 );
	}
}
